@@ -12,11 +12,10 @@ XPath 2.0 implementation - part 1 (parser class and symbols)
 """
 from abc import ABCMeta
 import locale
-from collections.abc import MutableSequence
+from collections.abc import Callable, MutableMapping
 from urllib.parse import urlparse
-from typing import cast, Any, ClassVar, Dict, List, Optional, Tuple, Type, Union
+from typing import cast, Any, ClassVar, Optional, Union
 
-from elementpath._typing import Callable, MutableMapping
 from elementpath.aliases import NamespacesType, NargsType
 from elementpath.helpers import upper_camel_case, is_ncname, ordinal
 from elementpath.exceptions import ElementPathError, ElementPathTypeError, \
@@ -74,7 +73,7 @@ class XPath2Parser(XPath1Parser):
     """
     version = '2.0'
 
-    DEFAULT_NAMESPACES: ClassVar[Dict[str, str]] = {
+    DEFAULT_NAMESPACES: ClassVar[dict[str, str]] = {
         'xml': XML_NAMESPACE,
         'xs': XSD_NAMESPACE,
         'fn': XPATH_FUNCTIONS_NAMESPACE,
@@ -93,8 +92,8 @@ class XPath2Parser(XPath1Parser):
         'schema-element', 'text', 'typeswitch',
     }
 
-    function_signatures: Dict[Tuple[QName, int], str] = XPath1Parser.function_signatures.copy()
-    namespaces: Dict[str, str]
+    function_signatures: dict[tuple[QName, int], str] = XPath1Parser.function_signatures.copy()
+    namespaces: dict[str, str]
     token: XPathToken
     next_token: XPathToken
 
@@ -107,8 +106,8 @@ class XPath2Parser(XPath1Parser):
                  xsd_version: Optional[str] = None,
                  schema: Optional[AbstractSchemaProxy] = None,
                  base_uri: Optional[str] = None,
-                 variable_types: Optional[Dict[str, str]] = None,
-                 document_types: Optional[Dict[str, str]] = None,
+                 variable_types: Optional[dict[str, str]] = None,
+                 document_types: Optional[dict[str, str]] = None,
                  collection_types: Optional[NamespacesType] = None,
                  default_collection_type: str = 'node()*') -> None:
 
@@ -199,7 +198,7 @@ class XPath2Parser(XPath1Parser):
             return f"{repr_string}{', '.join(args)})"
         return f"{repr_string}, {', '.join(args)})"
 
-    def __getstate__(self) -> Dict[str, Any]:
+    def __getstate__(self) -> dict[str, Any]:
         state = self.__dict__.copy()
         state.pop('symbol_table', None)
         state.pop('tokenizer', None)
@@ -242,11 +241,17 @@ class XPath2Parser(XPath1Parser):
 
     @classmethod
     def constructor(cls, symbol: str, bp: int = 90, nargs: NargsType = 1,
-                    sequence_types: Union[Tuple[()], Tuple[str, ...], List[str]] = (),
-                    label: Union[str, Tuple[str, ...]] = 'constructor function') \
+                    sequence_types: Union[tuple[()], tuple[str, ...], list[str]] = (),
+                    label: Union[str, tuple[str, ...]] = 'constructor function') \
             -> Callable[[Callable[..., Any]], Callable[..., Any]]:
-        """Creates a constructor token class."""
+        """
+        Statically creates a constructor token class, that is registered in the globals
+        of the module where the method is called.
+        """
         def nud_(self: XPathConstructor) -> XPathConstructor:
+            if not self.parser.parse_arguments:
+                return self
+
             try:
                 self.parser.advance('(')
                 self[0:] = self.parser.expression(5),
@@ -262,7 +267,7 @@ class XPath2Parser(XPath1Parser):
                 return self
 
         def evaluate_(self: XPathConstructor, context: Optional[XPathContext] = None) \
-                -> Union[List[None], AtomicType]:
+                -> Union[list[None], AtomicType]:
             if self.context is not None:
                 context = self.context
 
@@ -300,8 +305,8 @@ class XPath2Parser(XPath1Parser):
         return bind
 
     def schema_constructor(self, atomic_type_name: str, bp: int = 90) \
-            -> Type[XPathFunction]:
-        """Registers a token class for a schema atomic type constructor function."""
+            -> type[XPathFunction]:
+        """Dynamically registers a token class for a schema atomic type constructor function."""
         if atomic_type_name in (XSD_ANY_ATOMIC_TYPE, XSD_NOTATION):
             raise xpath_error('XPST0080')
 
@@ -317,7 +322,7 @@ class XPath2Parser(XPath1Parser):
             return self_
 
         def evaluate_(self_: XPathFunction, context: Optional[XPathContext] = None) \
-                -> Union[List[None], AtomicType]:
+                -> Union[list[None], AtomicType]:
             arg = self_.get_argument(context)
             if arg is None or self_.parser.schema is None:
                 return []
@@ -346,9 +351,8 @@ class XPath2Parser(XPath1Parser):
             '__return__': None
         }
         token_class = cast(
-            Type[XPathFunction], ABCMeta(token_class_name, (XPathFunction,), kwargs)
+            type[XPathFunction], ABCMeta(token_class_name, (XPathFunction,), kwargs)
         )
-        MutableSequence.register(token_class)
 
         if self.symbol_table is self.__class__.symbol_table:
             self.symbol_table = dict(self.__class__.symbol_table)
@@ -361,8 +365,8 @@ class XPath2Parser(XPath1Parser):
                           callback: Callable[..., Any],
                           name: Optional[str] = None,
                           prefix: Optional[str] = None,
-                          sequence_types: Tuple[str, ...] = (),
-                          bp: int = 90) -> Type[XPathFunction]:
+                          sequence_types: tuple[str, ...] = (),
+                          bp: int = 90) -> type[XPathFunction]:
         """Registers a token class for an external function."""
         import inspect
 
@@ -413,24 +417,23 @@ class XPath2Parser(XPath1Parser):
                     msg = f'function {qname.qname!r} is already registered'
                     raise ElementPathValueError(msg)
 
-                # Copy the token class before register the proxy token
+                # Move the token class before register the proxy token
                 self.symbol_table[f'{{{token_cls.namespace}}}{symbol}'] = token_cls
 
-            proxy_class_name = f'{upper_camel_case(qname.local_name)}ProxyToken'
+            token_class_name = f'{upper_camel_case(qname.local_name)}FunctionProxy'
             kwargs = {
-                'class_name': proxy_class_name,
+                'class_name': token_class_name,
                 'symbol': symbol,
+                'label': 'function',
                 'lbp': bp,
                 'rbp': bp,
                 '__module__': self.__module__,
-                '__qualname__': proxy_class_name,
+                '__qualname__': token_class_name,
                 '__return__': None
             }
-            proxy_class = cast(
-                Type[ProxyToken], ABCMeta(class_name, (ProxyToken,), kwargs)
+            self.symbol_table[symbol] = cast(
+                type[ProxyToken], ABCMeta(class_name, (ProxyToken,), kwargs)
             )
-            MutableSequence.register(proxy_class)
-            self.symbol_table[symbol] = proxy_class
 
         def evaluate_external_function(self_: XPathFunction,
                                        context: Optional[XPathContext] = None) -> Any:
@@ -492,10 +495,8 @@ class XPath2Parser(XPath1Parser):
                     )
 
         token_class = cast(
-            Type[XPathFunction], ABCMeta(class_name, (XPathFunction,), kwargs)
+            type[XPathFunction], ABCMeta(class_name, (XPathFunction,), kwargs)
         )
-        MutableSequence.register(token_class)
-
         self.symbol_table[lookup_name] = token_class
         self.tokenizer = None
         return token_class
@@ -503,27 +504,6 @@ class XPath2Parser(XPath1Parser):
     def is_schema_bound(self) -> bool:
         return self.schema is not None and 'symbol_table' in self.__dict__
 
-    def parse(self, source: str) -> XPathToken:
-        if self.tokenizer is None:
-            self.tokenizer = self.create_tokenizer(self.symbol_table)
-
-        root_token = super(XPath1Parser, self).parse(source)
-        if root_token.label in ('sequence type', 'function test'):
-            raise root_token.error('XPST0003', "not allowed in XPath expression")
-
-        try:
-            root_token.evaluate()  # Static context evaluation
-        except MissingContextError:
-            pass
-
-        if self.schema is not None:
-            # Tokens tree labeling with XSD types using a dynamic schema context
-            context = self.schema.get_context()
-            for _ in root_token.select(context):
-                pass
-
-        return root_token
-
     def check_variables(self, values: MutableMapping[str, Any]) -> None:
         if self.variable_types is None:
             return
@@ -557,11 +537,6 @@ XPath2Parser.unregister('starts-with')
 
 ###
 # Symbols
-XPath2Parser.register('then')
-XPath2Parser.register('else')
-XPath2Parser.register('in')
-XPath2Parser.register('return')
-XPath2Parser.register('satisfies')
 XPath2Parser.register('?')
 XPath2Parser.register('(:')
 XPath2Parser.register(':)')
